use std::sync::Arc;

use connection::SupervisorSender;
use jabber::JID;
use sqlx::SqlitePool;
use stanza::roster;
use tokio::{
    sync::{mpsc, oneshot, Mutex},
    task::JoinSet,
};

use crate::connection::write::WriteHandle;
use crate::connection::{SupervisorCommand, SupervisorHandle};
use crate::error::Error;

mod connection;
mod error;

pub struct Luz {
    receiver: mpsc::Receiver<CommandMessage>,
    jid: Arc<Mutex<JID>>,
    // TODO: use a dyn passwordprovider trait to avoid storing password in memory
    password: String,
    connected: Arc<Mutex<Option<(WriteHandle, SupervisorHandle)>>>,
    db: SqlitePool,
    sender: mpsc::Sender<UpdateMessage>,
    /// if connection was shut down due to e.g. server shutdown, supervisor must be able to mark client as disconnected
    connection_supervisor_shutdown: oneshot::Receiver<()>,
    // TODO: will need to have an auto reconnect state as well (e.g. in case server shut down, to try and reconnect later)
    // TODO: will grow forever at this point, maybe not required as tasks will naturally shut down anyway?
    tasks: JoinSet<()>,
}

impl Luz {
    fn new(
        receiver: mpsc::Receiver<CommandMessage>,
        jid: Arc<Mutex<JID>>,
        password: String,
        connected: Arc<Mutex<Option<(WriteHandle, SupervisorHandle)>>>,
        connection_supervisor_shutdown: oneshot::Receiver<()>,
        db: SqlitePool,
        sender: mpsc::Sender<UpdateMessage>,
    ) -> Self {
        Self {
            jid,
            password,
            connected,
            db,
            receiver,
            sender,
            tasks: JoinSet::new(),
            connection_supervisor_shutdown,
        }
    }

    async fn run(mut self) {
        loop {
            tokio::select! {
                _ = &mut self.connection_supervisor_shutdown => {
                    *self.connected.lock().await = None
                }
                Some(msg) = self.receiver.recv() => {
                    // TODO: consider separating disconnect/connect and commands apart from commandmessage
                    // TODO: dispatch commands separate tasks
                    match msg {
                        CommandMessage::Connect => {
                            match self.connected.lock().await.as_ref() {
                                Some(_) => {
                                    self.sender
                                        .send(UpdateMessage::Error(Error::AlreadyConnected))
                                        .await;
                                }
                                None => {
                                    let mut jid = self.jid.lock().await;
                                    let mut domain = jid.domainpart.clone();
                                    // TODO: check what happens upon reconnection with same resource (this is probably what one wants to do and why jid should be mutated from a bare jid to one with a resource)
                                    let streams_result =
                                        jabber::connect_and_login(&mut jid, &self.password, &mut domain)
                                            .await;
                                    match streams_result {
                                        Ok(s) => {
                                            let (shutdown_send, shutdown_recv) = oneshot::channel::<()>();
                                            let (writer, supervisor) = SupervisorHandle::new(
                                                s,
                                                self.sender.clone(),
                                                self.db.clone(),
                                                shutdown_send,
                                            );
                                            self.connection_supervisor_shutdown = shutdown_recv;
                                            *self.connected.lock().await = Some((writer, supervisor));
                                        }
                                        Err(e) => {
                                            self.sender.send(UpdateMessage::Error(e.into()));
                                        }
                                    }
                                }
                            };
                        }
                        CommandMessage::Disconnect => match self.connected.lock().await.as_mut() {
                            None => {
                                self.sender
                                    .send(UpdateMessage::Error(Error::AlreadyDisconnected))
                                    .await;
                            }
                            mut c => {
                                if let Some((_write_handle, supervisor_handle)) = c.take() {
                                    let _ = supervisor_handle.send(SupervisorCommand::Disconnect).await;
                                } else {
                                    unreachable!()
                                };
                            }
                        },
                        _ => {
                            match self.connected.lock().await.as_ref() {
                                Some((w, s)) => self.tasks.spawn(msg.handle_online(
                                    w.clone(),
                                    s.sender(),
                                    self.jid.clone(),
                                    self.db.clone(),
                                    self.sender.clone(),
                                    // TODO: iq hashmap
                                )),
                                None => self.tasks.spawn(msg.handle_offline(
                                    self.jid.clone(),
                                    self.db.clone(),
                                    self.sender.clone(),
                                )),
                            };
                        }
                    }
                },
                else => break,
            }
        }
    }
}

impl CommandMessage {
    pub async fn handle_offline(
        mut self,
        jid: Arc<Mutex<JID>>,
        db: SqlitePool,
        sender: mpsc::Sender<UpdateMessage>,
    ) {
        todo!()
    }

    pub async fn handle_online(
        mut self,
        write_handle: WriteHandle,
        supervisor_control: SupervisorSender,
        // TODO: jid could lose resource by the end
        jid: Arc<Mutex<JID>>,
        db: SqlitePool,
        sender: mpsc::Sender<UpdateMessage>,
    ) {
        todo!()
    }
}

// TODO: separate sender and receiver, store handle to Luz process to ensure dropping
pub struct LuzHandle {
    sender: mpsc::Sender<CommandMessage>,
    receiver: mpsc::Receiver<UpdateMessage>,
}

impl LuzHandle {
    pub fn new(jid: JID, password: String, db: SqlitePool) -> Self {
        let (command_sender, command_receiver) = mpsc::channel(20);
        let (update_sender, update_receiver) = mpsc::channel(20);
        // might be bad, first supervisor shutdown notification oneshot is never used (disgusting)
        let (sup_send, sup_recv) = oneshot::channel();

        let actor = Luz::new(
            command_receiver,
            Arc::new(Mutex::new(jid)),
            password,
            Arc::new(Mutex::new(None)),
            sup_recv,
            db,
            update_sender,
        );
        tokio::spawn(async move { actor.run().await });

        Self {
            sender: command_sender,
            receiver: update_receiver,
        }
    }
}

pub enum CommandMessage {
    Connect,
    Disconnect,
    /// gets the roster. if offline, retreives cached version from database. should be stored in application memory.
    GetRoster,
    SendMessage(JID, String),
}

pub enum UpdateMessage {
    Error(Error),
    Roster(Vec<roster::Item>),
}
